# encoding=utf-8
"""
3D visualization tools for geometries
"""
# %%
import copy
import warnings
import itertools

import torch

from torchgdm.constants import COLORS_DEFAULT
from torchgdm.tools.misc import to_np
from torchgdm import tools


# global color handling
COLOR_ITERATOR = itertools.cycle(COLORS_DEFAULT)
LEGEND_ENTRIES_LOOKUP = dict()


# reset color iterator and materials lookup
def _reset_color_iterator():
    global COLOR_ITERATOR, LEGEND_ENTRIES_LOOKUP
    COLOR_ITERATOR = itertools.cycle(COLORS_DEFAULT)
    LEGEND_ENTRIES_LOOKUP = dict()


def _return_next_color():
    return next(COLOR_ITERATOR)


_reset_color_iterator()


def _generate_legend(pl):
    global LEGEND_ENTRIES_LOOKUP

    if len(LEGEND_ENTRIES_LOOKUP) != 0:
        labels = LEGEND_ENTRIES_LOOKUP.keys()
    else:
        labels = []

    pv_labels = []
    for i_s, label in enumerate(labels):
        legend_dict = LEGEND_ENTRIES_LOOKUP[label]
        pv_labels.append([label, legend_dict["fc"]])

    pl.add_legend(labels=pv_labels, bcolor="w", face=None)

    return pl.legend


def _plot_structure_discretized(
    struct,
    scale=1.0,
    color="auto",
    show_grid=True,
    legend=True,
    alpha=1.0,
    show="auto",
    pl=None,
    reset_color_cycle=True,
    struct_dim="auto",
    infinite_axis_length=1000,
    **kwargs,
):
    global COLOR_ITERATOR, LEGEND_ENTRIES_LOOKUP
    import numpy as np
    import pyvista as pv

    if "projection" in kwargs:
        kwargs.pop("projection")

    if struct_dim.lower() == "auto":
        struct_dim = struct.n_dim
    else:
        if struct_dim not in [2, 3]:
            raise ValueError(
                "3D geometry plot: Structure dimension must be either 2 or 3."
            )

    if reset_color_cycle:
        _reset_color_iterator()

    # get mesh positions and step sizes, cut in multi-materials
    pos = to_np(struct.positions)
    step = to_np(struct.step)

    if type(show) == str:
        if show.lower() == "auto" and pl is None:
            show = True
        else:
            False

    if pl is None:
        pl = pv.Plotter()

    if color == "auto":
        # colors = COLORS_DEFAULT
        diff_mat_names = [s.__name__ for s in struct.materials]
        mat_names = np.array(diff_mat_names)

        different_materials = np.unique(mat_names)

        mat_pos_subset_idx = []
        for pos_single_mat in different_materials:
            mat_pos_subset_idx.append(
                np.arange(len(mat_names))[mat_names == pos_single_mat]
            )
    else:
        different_materials = ["struct. id:{}".format(struct.id)]
        mat_pos_subset_idx = [np.arange(len(pos))]  # all pos
        # colors = [color]

    # the actual plot
    mesh_list = []
    for i_s, pos_idx in enumerate(mat_pos_subset_idx):
        pos_mat = pos[pos_idx]
        steplist_mat = step[pos_idx]

        # 3D plot
        if struct_dim == 3:
            pts = pv.PolyData(pos_mat)
            pts["steps"] = steplist_mat
            pts.set_active_scalars("steps")
            mesh_list.append(
                pts.glyph(geom=pv.Cube(), scale="steps", factor=scale, orient=False)
            )
        # 2D plot (plot very long "infinite" axis)
        elif struct_dim == 2:
            _b = []
            for i, _s in enumerate(steplist_mat):
                pts = pv.PolyData([pos_mat[i]])
                _b.append(
                    pts.glyph(
                        geom=pv.Box(
                            bounds=(
                                -0.5 * _s,
                                0.5 * _s,
                                -0.5,
                                infinite_axis_length,
                                -0.5 * _s,
                                0.5 * _s,
                            ),
                            level=0,
                        ),
                        orient=False,
                    )
                )
            mesh_list.append(pv.merge(_b))

    for i_s, mesh in enumerate(mesh_list):
        # chose color
        if color == "auto":
            # if material already used, re-use its color. otherwise use next color
            if different_materials[i_s] in LEGEND_ENTRIES_LOOKUP:
                _col = LEGEND_ENTRIES_LOOKUP[different_materials[i_s]]["color"]
            else:
                _col = next(COLOR_ITERATOR)
        elif type(color) in [list, tuple]:
            _col = color[i_s]
        else:
            _col = color

        # add legend entry:
        LEGEND_ENTRIES_LOOKUP[different_materials[i_s]] = dict(
            color=_col,
            fc=_col,
            ec=_col,
            marker="s",
            markersize=10,
        )

        label = different_materials[i_s]
        pl.add_mesh(
            mesh,
            color=_col,
            show_edges=show_grid,
            edge_color="black",
            line_width=0.5,
            opacity=alpha,
            edge_opacity=alpha,
            label=label,
        )

    if legend:
        _generate_legend(pl)

    if show:
        pl.show()
        pl.camera.azimuth = 180.0  # convention: x-axis to the right

    return mesh_list


def _plot_structure_eff_3dpola(
    struct,
    scale=1.0,
    center_marker_scale=10,
    color="auto",
    sphere_style="wireframe",
    color_sphere="auto",
    theta_resolution=20,
    phi_resolution=20,
    alpha=0.1,
    show_grid=True,
    color_grid="auto",
    alpha_grid=0.5,
    show="auto",
    pl=None,
    legend=False,
    reset_color_cycle=True,
    struct_dim="auto",
    infinite_axis_length=1000,
    **kwargs,
):
    global COLOR_ITERATOR, LEGEND_ENTRIES_LOOKUP
    from torchgdm import tools
    import numpy as np
    import pyvista as pv

    if "projection" in kwargs:
        kwargs.pop("projection")

    if struct_dim.lower() == "auto":
        struct_dim = struct.n_dim
    else:
        if struct_dim not in [2, 3]:
            raise ValueError(
                "3D geometry plot: Structure dimension must be either 2 or 3."
            )

    if reset_color_cycle:
        _reset_color_iterator()

    if type(show) == str:
        if show.lower() == "auto" and pl is None:
            show = True
        else:
            show = False

    if pl is None:
        pl = pv.Plotter()

    # legend label based on structure ID
    legend_label = "eff.pola. (id: {})".format(struct.id)

    # automatic next color
    if color == "auto":
        if legend_label in LEGEND_ENTRIES_LOOKUP:
            _col = LEGEND_ENTRIES_LOOKUP[legend_label]["color"]
        else:
            _col = next(COLOR_ITERATOR)
    else:
        _col = color

    if color_grid == "auto":
        color_grid = _col
    if color_sphere == "auto":
        color_sphere = _col

    # add / update legend entry
    LEGEND_ENTRIES_LOOKUP[legend_label] = dict(
        color=_col,
        fc=color_sphere,
        ec=_col,
        marker="o",
        markersize=10,
    )

    # geometry data to numpy
    pos_a = to_np(struct.positions)
    enclosing_radius = to_np(struct.step) / 2.0

    # create the actual plot: iterate over polarizabilities
    mesh_sphere_list = []
    mesh_center_list = []
    mesh_fullgeo_list = []
    for i, pos in enumerate(pos_a):
        r = enclosing_radius[i]
        if len(struct.full_geometries) > 0:
            _geo = struct.full_geometries[i]
            pos_mesh = to_np(_geo)
            step_mesh = to_np(tools.geometry.get_step_from_geometry(_geo))
        else:
            pos_mesh = []
            step_mesh = 0.0

        # 3D: spheres
        if struct_dim == 3:
            # plot enclosing sphere
            mesh_sphere_list.append(
                pv.Sphere(
                    r + step_mesh / 2.0,
                    pos,
                    theta_resolution=theta_resolution,
                    phi_resolution=phi_resolution,
                )
            )

            # center pos. "marker" sphere
            mesh_center_list.append(pv.Sphere(center_marker_scale, pos))

            # full geometry mesh
            pts = pv.PolyData(pos_mesh)
            pts["steps"] = np.ones(len(pos_mesh)) * step_mesh
            pts.set_active_scalars("steps")

            mesh_fullgeo_list.append(
                pts.glyph(geom=pv.Cube(), scale="steps", factor=scale, orient=False)
            )
        # 2D plot (plot very long "infinite" axis)
        elif struct_dim == 2:
            _p = pos.copy()
            _p[1] += infinite_axis_length / 2
            # plot enclosing circle
            mesh_sphere_list.append(
                pv.Cylinder(
                    center=_p,
                    direction=[0, 1, 0],
                    radius=r + step_mesh / 2.0,
                    height=infinite_axis_length,
                    resolution=theta_resolution,
                )
            )

            # center pos. "marker" cylinder
            mesh_center_list.append(
                pv.Cylinder(
                    center=_p,
                    direction=[0, 1, 0],
                    radius=center_marker_scale,
                    height=infinite_axis_length,
                )
            )

            # full geometry mesh
            if len(struct.full_geometries) > 0:
                _b = []
                _s = step_mesh
                for i, _pos in enumerate(pos_mesh):
                    pts = pv.PolyData([_pos])
                    _b.append(
                        pts.glyph(
                            geom=pv.Box(
                                bounds=(
                                    -0.5 * _s,
                                    0.5 * _s,
                                    -0.5,
                                    infinite_axis_length,
                                    -0.5 * _s,
                                    0.5 * _s,
                                ),
                                level=0,
                            ),
                            orient=False,
                        )
                    )
                mesh_fullgeo_list.append(pv.merge(_b))

    # plot enclosing sphere wireframe
    for i_s, mesh in enumerate(mesh_sphere_list):
        pl.add_mesh(
            mesh,
            color=color_sphere,
            show_edges=False,
            line_width=0.5,
            edge_opacity=alpha,
            opacity=alpha,
            style=sphere_style,
        )

    # plot dipole position
    for i_s, mesh in enumerate(mesh_center_list):
        pl.add_mesh(mesh, color=_col)

    # optionally plot the replaced full geometry mesh
    if show_grid and len(struct.full_geometries) > 0:
        for i_s, mesh in enumerate(mesh_fullgeo_list):
            pl.add_mesh(
                mesh,
                color=color_grid,
                show_edges=True,
                edge_color="black",
                line_width=0.5,
                opacity=alpha_grid * alpha,
                edge_opacity=alpha_grid * alpha * 0.1,
            )

    # optional legend additions
    if legend:
        _generate_legend(pl)

    if show:
        pl.show()
        pl.camera.azimuth = 180.0  # convention: x-axis to the right

    return mesh_sphere_list, mesh_center_list, mesh_fullgeo_list


def structure(
    struct,
    color="auto",
    scale=1,
    legend=True,
    reset_color_cycle=True,
    pl=None,
    show="auto",
    struct_dim="auto",
    infinite_axis_length=1000,
    **kwargs,
):
    """plot structure in 3D

    Args:
        struct (struct): simulation or structure
        color (str, optional): color of the structure. If "auto", cycle through materials. Defaults to "auto".
        scale (int, optional): global scaling factor. Defaults to 1.
        legend (bool, optional): Whether to show a material legend. Defaults to True.
        reset_color_cycle (bool, optional): restart color cycle relative to former plots. Defaults to True.
        pl (pvista plotter, optional): pvista plotter to plot in. If not given, start new one. Defaults to None.
        show (str, optional): whether to show the figure. If auto, will only show if a new plotter is created. Defaults to "auto".
        struct_dim (str, optional): structure dimension: 2 or 3. in 2D structures, plot long cylinders along the infite axis. Defaults to "auto".
        infinite_axis_length (int, optional): absolute length of infinite axis in 2D plots (in nm!). Defaults to 1000.

    Raises:
        ValueError: incorrect configurations

    Returns:
        pyvista meshes
    """
    import pyvista as pv
    from torchgdm.simulation import SimulationBase, Simulation
    from torchgdm.struct.volume.pola import StructDiscretized3D
    from torchgdm.struct.volume.pola import StructDiscretizedCubic3D
    from torchgdm.struct.volume.pola import StructDiscretizedHexagonal3D
    from torchgdm.struct.point.pola import StructEffPola3D

    if reset_color_cycle:
        _reset_color_iterator()

    # got a structure instance:
    if (
        issubclass(type(struct), StructDiscretized3D)
        or issubclass(type(struct), StructDiscretizedCubic3D)
        or issubclass(type(struct), StructDiscretizedHexagonal3D)
    ):
        pl_res = _plot_structure_discretized(
            struct,
            color=color,
            scale=scale,
            show=show,
            pl=pl,
            struct_dim=struct_dim,
            infinite_axis_length=infinite_axis_length,
            **kwargs,
        )
    elif issubclass(type(struct), StructEffPola3D):
        pl_res = _plot_structure_eff_3dpola(
            struct,
            color=color,
            scale=scale,
            show=show,
            pl=pl,
            struct_dim=struct_dim,
            infinite_axis_length=infinite_axis_length,
            **kwargs,
        )
    elif issubclass(type(struct), Simulation) or issubclass(
        type(struct), SimulationBase
    ):
        # -- prep
        sim = struct

        if type(show) == str:
            if show.lower() == "auto" and pl is None:
                show = True
            else:
                show = False

        if pl is None:
            pl = pv.Plotter()

        # -- call all structure's plot functions
        pl_res = []  # collect results
        for i_s, _st in enumerate(sim.structures):
            pl_res.append(
                _st.plot3d(
                    color=color,
                    scale=scale,
                    pl=pl,
                    legend=False,
                    reset_color_cycle=False,
                    show=False,
                    struct_dim=struct_dim,
                    infinite_axis_length=infinite_axis_length,
                    **kwargs,
                )
            )

        # -- finalize: config global plot
        if legend:
            _generate_legend(pl)
        if show:
            pl.show()
            pl.camera.azimuth = 180.0  # convention: x-axis to the right
    else:
        raise ValueError("Unknown structure input")

    if reset_color_cycle:
        _reset_color_iterator()

    return pl_res



# encoding=utf-8
"""3D homogeneous environment and illuminations

Classes
-------

along with the main environment definition, compatible illuminations are defined:

.. currentmodule:: torchgdm.env.freespace_3d

.. autosummary::
   :toctree: generated/

   EnvHomogeneous3D
   NullField
   PlaneWave
   GaussianParaxial
   ElectricDipole
   MagneticDipole

"""
from . import inc_fields

from .dyads import EnvHomogeneous3D

from .inc_fields import NullField
from .inc_fields import PlaneWave
from .inc_fields import GaussianParaxial
from .inc_fields import ElectricDipole
from .inc_fields import MagneticDipole


# -*- coding: utf-8 -*-
"""batch processing tools: 

   - general batched evaluations (to limit memory usage)
   - multi-wavelength evaluation of spectra
   - tools for multi-illumination batched calculations
   
"""
# %%
import warnings
import gc
import typing

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device
from torchgdm.tools.misc import tqdm, test_illumination_field_config_exists


def batched(
    batch_kwarg: str,
    arg_position: int = None,
    out_dim_batch: int = 0,
    default_batch_size: int = 256,
    title="batched evaluation",
):
    """decorator to wrap a function for mini-batch evaluation

    Notes:
        - useful to avoid memory problems in large calculations
        - batched input parameter must be specified as kwarg.
        - func must return a torch tensor, a tuple of such or a dict containing such

    Args:
        batch_kwarg (str): kwarg of batched function that receives multiple configurations, for which batches are created
        arg_position (int, optional): position of the argument that receives a list for which batches are created. Defaults to None.
        out_dim_batch (int, optional): concatenation dimension of the original function's output. Defaults to 0.
        default_batch_size (int, optional): Defaults to 256.
        title (str, optional): title to show for the optional progress bar. Defaults to "batched evaluation".
    """

    def wrap(func):
        def batched_func(
            *args,
            batch_size=default_batch_size,
            progress_bar=False,
            pgr_bar_title=None,
            **kwargs,
        ):
            if pgr_bar_title is None:
                pgr_bar_title = title
            # get batched argument, try kwarg first. If not given, try positional arg
            try:
                d = kwargs[batch_kwarg]
            except KeyError as e:
                if arg_position is None:
                    raise e
                else:
                    d = args[arg_position]
                    args = list(
                        args
                    )  # hacky way to remove the batched arg. as it will be passed as kwarg.
                    args.pop(arg_position)
            N = len(d)

            # --- loop over batches
            def loop_batches(batch_size):
                results = []
                if progress_bar:
                    iterator = tqdm(range(0, N, batch_size), title=pgr_bar_title)
                else:
                    iterator = range(0, N, batch_size)
                # if progress_bar in kwargs:
                #     kwargs.pop("progress_bar")

                for i in iterator:
                    kwargs[batch_kwarg] = d[i : i + batch_size]
                    result = func(*args, **kwargs)

                    results.append(result)
                return results

            try:
                results = loop_batches(batch_size=batch_size)
            except torch.cuda.OutOfMemoryError as e:
                if batch_size > 1:
                    warnings.warn(
                        "CUDA out of memory exception occurred. "
                        + "Current batch_size={}. ".format(batch_size)
                        + "Re-trying with batch_size=1. "
                        + "This may be inefficient, and you may want "
                        + "to try setting `batch_size` yourself."
                    )
                    results = loop_batches(batch_size=1)
                else:
                    print(
                        "Out of CUDA memory. Try to reduce `batch_size` (currently, batch_size={})!".format(
                            batch_size
                        )
                    )
                    raise e

            # --- concatenate output batches
            # case multiple outputs
            if type(results[0]) == tuple:
                return [
                    torch.cat([res[i_out] for res in results], dim=out_dim_batch)
                    for i_out in range(len(results[0]))
                ]

            # case dictionary output
            if type(results[0]) == dict:
                # print("WARNING: NOT YET TESTED!! DOES IT WORK??")
                dict_out = dict()
                for k in results[0]:
                    dict_out[k] = torch.cat(
                        [_d[k] for _d in results], dim=out_dim_batch
                    )
                return dict_out

            # case single torch tensor outputs
            else:
                return torch.cat(results, dim=out_dim_batch)

        return batched_func

    return wrap


# --- helper for wavelength batched evaluation
def calc_spectrum(
    sim,
    func,
    illumination_index=None,
    progress_bar=True,
    progress_bar_title="spectrum",
    **kwargs,
):
    """calculate spectrum for all wavelengths defined in simulation

    kwargs are passed to `func`

    Args:
        sim (`torchgdm.Simulation`): simulation (`sim.run()` needs to have been executed before)
        func (`typing.Callable`): function to call with simulation. Needs to accept following kwargs: `sim`, `wavelength` and `illumination_index`
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
        progress_bar (bool): Whether to show a progress bar. Defaults to True

    Returns:
        dict: spectra results
         - contains spectra-lists with results returned by func
         - contains wavelengths
         - contains illumination field configs
    """
    wavelengths = sim.wavelengths

    # loop through wavelengths, add results to dict
    spectrum_results = {}
    for wl in tqdm(wavelengths, progress_bar, title=progress_bar_title):
        results = func(
            sim=sim,
            wavelength=wl,
            illumination_index=illumination_index,
            progress_bar=False,
            **kwargs,
        )
        for k in results:
            if k != "e0_config":
                if k not in spectrum_results:
                    spectrum_results[k] = []
                spectrum_results[k].append(results[k])

    # if `func` result is a single tensor, stack tensors of spectrum
    for k in spectrum_results:
        if type(results[k]) == torch.Tensor:
            spectrum_results[k] = torch.stack(spectrum_results[k])

    # add wavelengths and return dict containing all spectra
    if "wavelengths" not in spectrum_results:
        spectrum_results["wavelengths"] = wavelengths

    # add information on illumination field config
    spectrum_results = add_illuminations_info_to_dict(
        sim, spectrum_results, illumination_index=illumination_index
    )

    return spectrum_results


# --- helper for raster-scan type multiple illuminations batch evaluation
def calc_rasterscan(sim, func, wavelength, kw_r="r_focus", **kwargs):
    """calculate rasterscan, using all field configurations defined in the simulation

    kwargs are passed to `func`

    Args:
        sim (`torchgdm.Simulation`): simulation (`sim.run()` needs to have been executed before)
        func (`typing.Callable`): function to call with simulation. Needs to accept following kwargs: `sim`, `wavelength` and `illumination_index`
        wavelength (float): in nm
        kw_r (`str`): kwarg used for scan position of rasterscan

    Returns:
        dict: rasterscan results
         - contains wavelength
         - contains lists with rasterscan results at each pos returned by func
         - contains positions of illumination
    """
    # loop through illuminations
    pos = []
    for i, e_inc in enumerate(sim.illumination_fields):
        pos.append(getattr(e_inc, kw_r))
    pos = torch.stack(pos).to(device=sim.device, dtype=DTYPE_FLOAT)
    rs_results = dict(positions=pos)

    # evaluate and loop through results
    all_conf_res = func(sim=sim, wavelength=wavelength, **kwargs)

    for key in all_conf_res:
        if key != "e0_config":
            if key not in rs_results:
                rs_results[key] = []
            rs_results[key].append(all_conf_res[key])

    # if `func` result is a single tensor, stack tensors of spectrum
    for key in rs_results:
        if key not in ["positions", "wavelength"]:
            if type(all_conf_res[key]) == torch.Tensor:
                rs_results[key] = torch.stack(rs_results[key])

    # add working wavelength
    rs_results["wavelength"] = wavelength

    return rs_results


# --- helper for illumination field batch evaluation
def get_p_m_moments(sim, wavelength: float, illumination_index: int = None, **kwargs):
    """get self-consistent internal dipole moments at polarizable locations and their locations

    further kwargs are ignored

    Args:
        sim (`torchgdm.Simulation`): simulation (`sim.run()` needs to have been executed before)
        wavelength (float): in nm
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

    Returns:
        tuple: tuple containing 4x `torch.Tensor` with the dipole moments and their 3D positions (p, m, r_p and r_m)
    """
    # verify if config exists
    test_illumination_field_config_exists(
        sim=sim, wavelength=wavelength, illumination_index=illumination_index
    )
    # get dipole moments and their positions for selected illumination
    r_p, r_m = sim._get_polarizable_positions_separate_p_m()
    p = sim._get_polarizable_dipole_moments_p(wavelength)
    m = sim._get_polarizable_dipole_moments_m(wavelength)
    if illumination_index is not None:
        p = p[illumination_index].unsqueeze(0)
        m = m[illumination_index].unsqueeze(0)
    return p, m, r_p, r_m


def get_p_m_radiative_correction_prefactors(sim, wavelength: float, **kwargs):
    """get radiative absorption cross-section correction corrections at polarizable locations

    further kwargs are ignored

    Args:
        sim (`torchgdm.Simulation`): simulation (`sim.run()` needs to have been executed before)
        wavelength (float): in nm

    Returns:
        tuple: Two `torch.Tensor` with the radiative corrections for all polarizable electric and magnetic dipoles. of shape: (N_polarizable_p, 3) and (N_polarizable_m, 3)
    """
    idx_p, idx_m = sim._get_polarizable_positions_indices_separate_p_m()

    rad_corr_factor_p = torch.cat(
        [
            _s.get_radiative_correction_prefactor_p(wavelength, sim.environment)
            for _s in sim.structures
        ]
    )

    rad_corr_factor_m = torch.cat(
        [
            _s.get_radiative_correction_prefactor_m(wavelength, sim.environment)
            for _s in sim.structures
        ]
    )

    # keep only for polarizable locations
    rad_corr_p = torch.index_select(rad_corr_factor_p, 0, idx_p)
    rad_corr_m = torch.index_select(rad_corr_factor_m, 0, idx_m)

    return rad_corr_p, rad_corr_m


def get_e_hfields(sim, wavelength: float, illumination_index: int = None, **kwargs):
    """get self-consistent internal fields at polarizable locations and their locations

    further kwargs are ignored

    Args:
        sim (`torchgdm.Simulation`): simulation (`sim.run()` needs to have been executed before)
        wavelength (float): in nm
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

    Returns:
        tuple: tuple containing 4x `torch.Tensor` with the E and H-fields and their 3D positions (e, h, r_e and r_h)
    """
    # verify if config exists
    test_illumination_field_config_exists(
        sim=sim, wavelength=wavelength, illumination_index=illumination_index
    )
    # get dipole moments and their positions for selected illumination
    r_e, r_h = sim._get_polarizable_positions_separate_p_m()
    e = sim._get_polarizablefields_e(wavelength)
    h = sim._get_polarizablefields_h(wavelength)
    if illumination_index is not None:
        e = e[illumination_index].unsqueeze(0)
        h = h[illumination_index].unsqueeze(0)
    return e, h, r_e, r_h


def get_e0(
    sim,
    wavelength: float,
    r_probe: torch.Tensor,
    illumination_index: int = None,
    **kwargs,
):
    """get illumination E-field at `r`

    further kwargs are ignored

    Args:
        sim (`torchgdm.Simulation`): simulation (`sim.run()` needs to have been executed before)
        wavelength (float): in nm
        r (torch.Tensor): position(s) at which to evaluate incident field
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

    Returns:
        torch.Tensor: incident E-field(s)
    """
    e0 = torch.stack(
        [
            f_inc.get_efield(
                r_probe, wavelength=wavelength, environment=sim.environment
            )
            for f_inc in sim.illumination_fields
        ],
        dim=0,
    )
    if illumination_index is not None:
        e0 = e0[illumination_index].unsqueeze(0)
    return e0


def get_h0(
    sim,
    wavelength: float,
    r_probe: torch.Tensor,
    illumination_index: int = None,
    **kwargs,
):
    """get illumination H-field at `r`

    further kwargs are ignored

    Args:
        sim (`torchgdm.Simulation`): simulation (`sim.run()` needs to have been executed before)
        wavelength (float): in nm
        r (torch.Tensor): position(s) at which to evaluate incident field
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

    Returns:
        torch.Tensor: incident H-field(s)
    """
    h0 = torch.stack(
        [
            f_inc.get_hfield(
                r_probe, wavelength=wavelength, environment=sim.environment
            )
            for f_inc in sim.illumination_fields
        ],
        dim=0,
    )
    if illumination_index is not None:
        h0 = h0[illumination_index].unsqueeze(0)
    return h0


def get_k0_vec(
    sim,
    wavelength: float,
    r_probe: torch.Tensor,
    illumination_index: int = None,
    **kwargs,
):
    """get illumination wave vector at `r`

    Note: For a non-plane wave field, the scaled time-averaged Poynting vector is used, which is identical to the wave vector only in isotropic media.

    further kwargs are ignored

    Args:
        sim (`torchgdm.Simulation`): simulation (`sim.run()` needs to have been executed before)
        wavelength (float): in nm
        r (torch.Tensor): position(s) at which to evaluate incident field's wave vector
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

    Returns:
        torch.Tensor: wave vector(s) at `r`
    """
    k0_list = []
    for f_inc in sim.illumination_fields:
        try:
            propagation_vector = f_inc.k_vec_unit.real
            if len(propagation_vector.shape) == 1:
                propagation_vector = propagation_vector.unsqueeze(0)
        except AttributeError:
            warnings.warn(
                "Global wavevector not defined in illumination field class. "
                + "Using local time averaged Poynting vector. "
                + "Note that this is identical to the wave vector only in isotropic media."
            )
            e0 = f_inc.get_efield(
                r_probe, wavelength=wavelength, environment=sim.environment
            )
            h0 = f_inc.get_hfield(
                r_probe, wavelength=wavelength, environment=sim.environment
            )
            propagation_vector = torch.cross(torch.conj(e0), h0, dim=-1).real

        # normalize to wavenumber
        k0_unit = propagation_vector / torch.linalg.norm(propagation_vector)
        k0 = k0_unit * 2 * torch.pi / wavelength
        k0_list.append(k0)

    k0_list = torch.stack(k0_list, dim=0).to(dtype=DTYPE_FLOAT, device=sim.device)
    if illumination_index is not None:
        k0_list = k0_list[illumination_index].unsqueeze(0)
    return k0_list


def get_illuminations_info(sim, illumination_index: int = None, **kwargs):
    """get dict with configuration info of the illumination(s)

    further kwargs are ignored
    """
    e0_info = [e_inc.get_info() for e_inc in sim.illumination_fields]

    if illumination_index is not None:
        e0_info = [e0_info[illumination_index]]

    return e0_info


def add_illuminations_info_to_dict(
    sim, data_dict, illumination_index: int = None, **kwargs
):
    """add illumination information to `data_dict` as key 'e0_config'

    further kwargs are ignored
    """
    if "e0_config" not in data_dict:
        data_dict["e0_config"] = get_illuminations_info(
            sim=sim, illumination_index=illumination_index
        )
    else:
        warnings.warn("dict already contains `e0_config` key. Do nothing.")
    return data_dict



# -*- coding: utf-8 -*-
"""torchgdm postprocessing modules

.. currentmodule:: torchgdm.postproc

.. autosummary::
   :toctree: generated/

   crosssect
   fields
   multipole
   green

"""
from . import crosssect
from . import fields
from . import multipole
from . import green


# -*- coding: utf-8 -*-
"""2D visualization tools based on matplotlib

Requires `matplotlib`. Install with:

.. code-block:: bash

     pip install matplotlib


Contains following plotting tools:

    
.. autosummary::
   :toctree: generated/
    
    structure
    contour
    scalarfield
    field_intensity
    field_amplitude
    vectorfield
    vectorfield_inside
    streamlines_energy_flux

"""
from . import geo2d, vec2d, scalar2d

from .geo2d import _reset_color_iterator

from .geo2d import structure, contour
from .scalar2d import _scalarfield as scalarfield
from .scalar2d import field_intensity, field_amplitude
from .vec2d import vectorfield, vectorfield_inside, streamlines_energy_flux



# -*- coding: utf-8 -*-
"""material optical properties

.. currentmodule:: torchgdm.materials

Classes
-------

.. autosummary::
   :toctree: generated/
   :recursive:

    MatConstant
    MatDatabase
    MatTiO2
    MaterialBase


Functions
---------

.. autosummary::
   :toctree: generated/

    list_available_materials

"""
from .base_classes import MaterialBase

from .hardcoded import MatConstant
from .hardcoded import MatTiO2

from .tabulated import list_available_materials
from .tabulated import MatDatabase
